import { WebAssemblyDecodeError } from './reader/errors.js';
import { readModuleRaw } from './reader/module-reader.js';
/**
 * Class representing a WebAssembly Module.
 */
export class Module {
    constructor(buffer) {
        this.importsByType = {
            function: [],
            global: [],
            memory: [],
            table: [],
        };
        /**
         * An array of module functions.
         */
        this.functions = [];
        /**
         * An array of global variables in the module.
         */
        this.globals = [];
        /**
         * An array of tables in the module.
         */
        this.tables = [];
        /**
         * An array of memory blocks in the module.
         */
        this.memories = [];
        /**
         * An array of module imports.
         */
        this.imports = [];
        /**
         * An array of module exports.
         */
        this.exports = [];
        const sections = Object.groupBy(readModuleRaw(buffer), (section) => section.type);
        function getSection(type) {
            return sections[type]?.[0];
        }
        const types = getSection('type')?.types ?? [];
        const processedImports = getSection('import')?.imports?.map((i) => mapImport(i, types));
        this.addImports(processedImports ?? []);
        const indices = getSection('function')?.indices ?? [];
        const functions = indices
            .map((i) => types?.[i])
            .filter(Boolean);
        this.addFunctions(functions);
        const globals = getSection('global')?.globals?.map((g) => g.type);
        this.addGlobals(globals ?? []);
        this.addMemories(getSection('memory')?.memories ?? []);
        this.addTables(getSection('table')?.tables ?? []);
        this.addExports(getSection('export')?.exports ?? []);
    }
    addFunction(funcType) {
        this.functions.push(mapFunction(funcType));
    }
    addFunctions(funcTypes) {
        this.functions = this.functions.concat(funcTypes.map(mapFunction));
    }
    addGlobal(global) {
        this.globals.push(mapGlobal(global));
    }
    addGlobals(globals) {
        this.globals = this.globals.concat(globals.map(mapGlobal));
    }
    addTable(table) {
        this.tables.push(mapTable(table));
    }
    addTables(tables) {
        this.tables = this.tables.concat(tables.map(mapTable));
    }
    addMemory(memory) {
        this.memories.push(mapMemory(memory));
    }
    addMemories(memories) {
        this.memories = this.memories.concat(memories.map(mapMemory));
    }
    addImport(imp) {
        this.imports.push(imp);
        this.importsByType[imp.target.kind].push(imp.target);
    }
    addImports(imports) {
        this.imports = this.imports.concat(imports);
        const grouped = Object.groupBy(imports.map((i) => i.target), (i) => i.kind);
        this.importsByType.function.push(...(grouped.function ?? []));
        this.importsByType.global.push(...(grouped.global ?? []));
        this.importsByType.memory.push(...(grouped.memory ?? []));
        this.importsByType.table.push(...(grouped.table ?? []));
    }
    addExport(exported) {
        this.exports.push(mapExport(this, exported));
    }
    addExports(exports) {
        this.exports = this.exports.concat(exports.map((e) => mapExport(this, e)));
    }
    /**
     * Resolves an export descriptor to its corresponding module symbol.
     *
     * @param descriptor - The export descriptor indicating the type and index of the export.
     * @return The resolved module symbol if found, otherwise undefined.
     */
    resolveExportDescriptor(descriptor) {
        const importsOfType = this.importsByType[descriptor.type];
        if (descriptor.index < importsOfType.length) {
            return importsOfType[descriptor.index];
        }
        const localIndex = descriptor.index - importsOfType.length;
        switch (descriptor.type) {
            case 'function':
                return this.functions[localIndex];
            case 'global':
                return this.globals[localIndex];
            case 'memory':
                return this.memories[localIndex];
            case 'table':
                return this.tables[localIndex];
        }
    }
}
function mapFunction(funcType) {
    return {
        kind: 'function',
        parametersTypes: funcType.parameterTypes,
        resultTypes: funcType.returnTypes,
    };
}
function mapGlobal(global) {
    return {
        kind: 'global',
        type: global.valueType,
        isMutable: global.isMutable,
    };
}
function mapMemory(memory) {
    return {
        kind: 'memory',
        minSize: memory.min,
        maxSize: memory.max,
    };
}
function mapTable(table) {
    return {
        kind: 'table',
        minSize: table.limits.min,
        maxSize: table.limits.max,
        elementType: table.elementType,
    };
}
/**
 * Resolves an import descriptor into a corresponding module symbol.
 *
 * @param descriptor - The descriptor that needs to be resolved into a module symbol.
 *                      It specifies the type and index of the symbol within the module.
 * @param types - Array of module types.
 * @returns The module symbol corresponding to the descriptor, or `undefined` if the descriptor
 *          cannot be resolved. The symbol can be a function, global, memory, or table, depending
 *          on the descriptor's type.
 */
function resolveImportDescriptor(descriptor, types) {
    switch (descriptor.type) {
        case 'function':
            const targetType = types[descriptor.index];
            if (!targetType) {
                throw new WebAssemblyDecodeError(`Could not decode import descriptor with index: ${descriptor.index}`);
            }
            return {
                kind: 'function',
                parametersTypes: targetType.parameterTypes,
                resultTypes: targetType.returnTypes,
            };
        case 'global':
            return mapGlobal(descriptor.global);
        case 'memory':
            return mapMemory(descriptor.memory);
        case 'table':
            return mapTable(descriptor.table);
    }
}
function mapImport(imp, types) {
    const target = resolveImportDescriptor(imp.descriptor, types);
    if (!target) {
        throw new WebAssemblyDecodeError(`Could not resolve descriptor ${JSON.stringify(imp.descriptor)}`);
    }
    return {
        kind: 'import',
        module: imp.module,
        name: imp.name,
        target,
    };
}
function mapExport(module, exp) {
    const target = module.resolveExportDescriptor(exp.descriptor);
    if (!target) {
        throw new WebAssemblyDecodeError(`Could not resolve descriptor ${JSON.stringify(exp.descriptor)}`);
    }
    return {
        kind: 'export',
        name: exp.name,
        target,
    };
}
